/*
 * asmloader.c
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PceNativeCall.h>
#include <Standalone.h>

#include "garux.h"

#define SET_TAG_WORD(off, value) *((unsigned long *)(0x10000100+((off)<<2)))=value
#define SET_TAG_WORD_BYTE(off, byte, value) *((char *)(0x10000100+((off)<<2)+(byte)))=value

STANDALONE_CODE_RESOURCE_ID (1);

//void startKernel();

unsigned long bootLinux(const void *emulStateP, char *userData68KP, Call68KFuncType *call68KFuncP) {
    unsigned long register i;
    unsigned long register j;
    unsigned long register krnlSrc;
    
    int byte = 0;
	char c, *cmdline;

    cmdline = (char*)userData68KP;
    krnlSrc=(unsigned long)userData68KP+256;

//io ports settings for uart3
    *((unsigned long*)0xfffe1044)=0xc9fbffff;
    *((unsigned long*)0xfffce004)=0x00003a0e;
    
//Reset for disable dma LCD
//irq reset
    *((unsigned long*)0xfffecb00)=0x0;
    *((unsigned long*)0xfffecb04)=0xffffffff;
//dmaGlobal reset
    *((unsigned short*)0xfffedb00)=0x0;
    *((unsigned short*)0xfffedc00)=0x8;


    // Disable FIQ, IRQ
    asm volatile ("mrs r0, cpsr");
    asm volatile ("orr r0, r0, #0xc0");
    asm volatile ("msr cpsr, r0");

    // Set Page table entries, On palm tt2 page table place in sram.
    *((unsigned long*)0x20010400)=0x10000c1a; //for 1MB SDRAM
    *((unsigned long*)0x20010404)=0x10100c1a; //for 2MB SDRAM

    // Let's flush Instruction and Data cache
    asm volatile ("mov r0, #0");
    asm volatile ("mcr p15, 0, r0, c7, c7, 0");       // flush v3/v4 cache
    asm volatile ("mcr p15, 0, r0, c8, c7, 0");       // flush v4 TLB

    // Wait a bit.
    for(i=0 ; i<1000000 ; i++);

    // Copy directly image to 0x10008000
    for(i=0 ; i<NUM_KERNEL_PAGES*PAGE_SIZE ; i+=4) {
        *((unsigned long *)(0x10008000+i))=*((unsigned long *)(krnlSrc+i));
    }

    // Set the ATags

    i=0;
    // ATAG_CORE
    SET_TAG_WORD(i++,0x00000002);
    SET_TAG_WORD(i++,0x54410001);

    // ATAG_MEM
    SET_TAG_WORD(i++,0x00000004);
    SET_TAG_WORD(i++,0x54410002);
    SET_TAG_WORD(i++,0x02000000);
    SET_TAG_WORD(i++,0x10000000);

     
    // ATAG_CMDLINE
	for (byte=0; cmdline[byte]; byte++);
    byte = (byte+3) & ~3;
    SET_TAG_WORD(i++,0x00000002+byte/4);
    SET_TAG_WORD(i++,0x54410009);
    byte = 0;
    do {
        c = *cmdline++;
        SET_TAG_WORD_BYTE(i,byte,c);
        if (++byte==4) { byte=0; i++; }
    } while (c);
    while (byte>0) {
        SET_TAG_WORD_BYTE(i,byte,0);
        if (++byte==4) { byte=0; i++; }
    }

    // ATAG_OMAP
    SET_TAG_WORD(i++,0x00000005);
    SET_TAG_WORD(i++,0x414f4d50);
    SET_TAG_WORD(i++,0x4f030005);
    SET_TAG_WORD(i++,0x030001c2);
    SET_TAG_WORD(i++,0x00000000);
   
    // ATAG_NONE
    SET_TAG_WORD(i++,0x00000000);
    SET_TAG_WORD(i++,0x00000000);


    // FIXME: We arbitrarily add 0x60 to pc and copy 512 bytes starting
    // from there. So if any code is modified between <CRITICAL>
    // and </CRITICAL>, the offset needs to change.
    // <CRITICAL>
    asm volatile ("beg_critical:");
    asm volatile ("add %0, pc, #0x30" : "=r" (j));
    for(i=0 ; i<512 ; i++) {
        *((char *)(0x10000000+i))=*((char *)(i+j));
    }

    asm volatile ("mov pc, #0x10000000");
    // </CRITICAL>
    asm volatile ("end_critical:");

    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");
    asm volatile ("nop");


    // Let's flush Instruction and Data cache
    asm volatile ("mov r0, #0");
    asm volatile ("mcr p15, 0, r0, c7, c7, 0");       // flush v3/v4 cache
    asm volatile ("mcr p15, 0, r0, c8, c7, 0");       // flush v4 TLB

    // Wait a bit.
    for(i=0 ; i<100000 ; i++);

    
    // Let's disable the MMU and disable data cache.
    asm volatile ("mrc p15, 0, r0, c1, c0, 0");
    asm volatile ("bic r0, r0, #0x00002300");         // clear bits 13, 9:8 (--V- --RS)
    asm volatile ("bic r0, r0, #0x00000087");         // clear bits 7, 2:0 (B--- -CAM)
    asm volatile ("orr r0, r0, #0x00001000");         // set bit 12 (I) I-Cache
    asm volatile ("mcr p15, 0, r0, c1, c0, 0");

    
asm volatile ("wait:");
    // Wait a bit.
    for(i=0 ; i<100000 ; i++);

asm volatile ("initrd_copy:");

    // copy the initrd to 0x10400000.
    for(i=0 ; i<SIZE_INITRD ; i++) {
        *((unsigned char *)(0x10400000+i))=*((unsigned char *)(0x10008000+SIZE_KERNEL+i));
    }

    // Wait a bit.
    for(i=0 ; i<10000000 ; i++);

    // Initializing before boot
    asm volatile ("mov r0, #0");
    asm volatile ("mov r1, #452"); //generic
//    asm volatile ("mov r1, #234");//innovator
    asm volatile ("mov r2, #0x10000000");
    asm volatile ("add r2, r2, #0x200");

asm volatile ("jmp_kernel:");
    // Jump to the kernel.
    asm volatile ("mov r3, #0x8000");
    asm volatile ("orr r3, r3, #0x10000000");
    asm volatile ("mov pc, r3");

    // We should never get here... Just to make the compiler happy.
    return 0;
}
